#ifndef AUBO_SDK_SOCKET_INTERFACE_H
#define AUBO_SDK_SOCKET_INTERFACE_H

#include <vector>
#include <memory>

#include <aubo/type_def.h>
#include <aubo/global_config.h>

namespace arcs {
namespace common_interface {

class ARCS_ABI_EXPORT Socket
{
public:
    Socket();
    virtual ~Socket();

    /**
     * Open TCP/IP ethernet communication socket
     *
     * Instruction
     */
    int socketOpen(const std::string &address, int port,
                   const std::string &socket_name = "socket_0");

    /**
     * Closes TCP/IP socket communication
     * Closes down the socket connection to the server.
     *
     * Instruction
     */
    int socketClose(const std::string &socket_name = "socket_0");

    /**
     * Reads a number of ascii formatted floats from the socket. A maximum
     * of 30 values can be read in one command.
     * A list of numbers read (list of floats, length=number+1)
     *
     * Instruction
     *
     * Result will be store in a register named reg_key. Use getFloatVec
     * to retrive data
     */
    int socketReadAsciiFloat(int number, const std::string &variable,
                             const std::string &socket_name = "socket_0");

    /**
     * Reads a number of 32 bit integers from the socket. Bytes are in
     * network byte order. A maximum of 30 values can be read in one
     * command.
     * A list of numbers read (list of ints, length=number+1)
     *
     * Instruction
     *
     * std::vector<int>
     */
    int socketReadBinaryInteger(int number, const std::string &variable,
                                const std::string &socket_name = "socket_0");

    /**
     * Reads a number of bytes from the socket. Bytes are in network byte
     * order. A maximum of 30 values can be read in one command.
     * A list of numbers read (list of ints, length=number+1)
     *
     * Instruction
     *
     * std::vector<char>
     */
    int socketReadByteList(int number, const std::string &variable,
                           const std::string &socket_name = "socket_0");

    /**
     * Reads all data from the socket and returns the data as a string.
     * Bytes are in network byte order.
     *
     * The optional parameters "prefix" and "suffix", can be used to express
     * what is extracted from the socket. The "prefix" specifies the start
     * of the substring (message) extracted from the socket. The data up to
     * the end of the "prefix" will be ignored and removed from the socket.
     * The "suffix" specifies the end of the substring (message) extracted
     * from the socket. Any remaining data on the socket, after the "suffix",
     * will be preserved. E.g. if the socket server sends a string
     * "noise>hello<", the controller can receive the "hello" by calling this
     * script function with the prefix=">" and suffix="<". By using the
     * "prefix" and "suffix" it is also possible send multiple string to the
     * controller at once, because the suffix defines where the message ends.
     * E.g. sending ">hello<>world<"
     *
     * Instruction
     *
     * std::string
     */
    int socketReadString(const std::string &variable,
                         const std::string &socket_name = "socket_0",
                         const std::string &prefix = "",
                         const std::string &suffix = "",
                         bool interpret_escape = false);

    /**
     *
     * Instruction
     * std::vector<char>
     */
    int socketReadAllString(const std::string &variable,
                            const std::string &socket_name = "socket_0");

    /**
     * Sends a byte to the server
     * Sends the byte <value> through the socket. Expects no response. Can
     * be used to send special ASCII characters; 10 is newline, 2 is start of
     * text, 3 is end of text.
     *
     * Instruction
     */
    int socketSendByte(char value, const std::string &socket_name = "socket_0");

    /**
     * Sends an int (int32_t) to the server
     * Sends the int <value> through the socket. Send in network byte order.
     * Expects no response
     *
     * Instruction
     */
    int socketSendInt(int value, const std::string &socket_name = "socket_0");

    /**
     * Sends a string with a newline character to the server - useful for
     * communicatin with the UR dashboard server
     * Sends the string <str> through the socket in ASCII coding. Expects no
     * response.
     *
     * Instruction
     */
    int socketSendLine(const std::string &str,
                       const std::string &socket_name = "socket_0");

    /**
     * Sends a string to the server
     * Sends the string <str> through the socket in ASCII coding. Expects no
     * response.
     *
     * Instruction
     */
    int socketSendString(const std::string &str,
                         const std::string &socket_name = "socket_0");
    int socketSendAllString(bool is_check, const std::vector<char> &str,
                            const std::string &socket_name = "socket_0");

protected:
    void *d_;
};
using SocketPtr = std::shared_ptr<Socket>;

// clang-format off
#define Socket_DECLARES                                                                \
    _INST(Socket, 3, socketOpen, address, port, socket_name)                               \
    _INST(Socket, 1, socketClose, socket_name)                                             \
    _FUNC(Socket, 3, socketReadAsciiFloat, number, variable, socket_name)                       \
    _FUNC(Socket, 3, socketReadBinaryInteger, number, variable, socket_name)                    \
    _FUNC(Socket, 3, socketReadByteList, number, variable, socket_name)                         \
    _FUNC(Socket, 5, socketReadString, variable, socket_name, prefix, suffix, interpret_escape) \
    _FUNC(Socket, 2, socketReadAllString, variable, socket_name)                                \
    _INST(Socket, 2, socketSendByte, value, socket_name)                                   \
    _INST(Socket, 2, socketSendInt, value, socket_name)                                    \
    _INST(Socket, 2, socketSendLine, str, socket_name)                                     \
    _INST(Socket, 2, socketSendString, str, socket_name)                                   \
    _INST(Socket, 3, socketSendAllString, is_check, str, socket_name)
// clang-format on

} // namespace common_interface
} // namespace arcs
#endif
