#include "aubo_sdk/rpc.h"

using namespace arcs::common_interface;
using namespace arcs::aubo_sdk;

void exampleRpyToQuat(RpcClientPtr impl)
{
    std::cout << std::endl << "欧拉角 -> 四元数 " << std::endl;

    std::vector<double> rpy;
    rpy.push_back(0.611);
    rpy.push_back(0.785);
    rpy.push_back(0.960);
    std::cout << "rx:" << rpy.at(0) << std::endl;
    std::cout << "ry:" << rpy.at(1) << std::endl;
    std::cout << "rz:" << rpy.at(2) << std::endl;

    auto robot_name = impl->getRobotNames().front();
    std::vector<double> quat = impl->getRobotInterface(robot_name)
                                   ->getRobotAlgorithm()
                                   ->rpyToQuaternion(rpy);

    if (4 == quat.size()) {
        std::cout << "w:" << quat.at(0) << std::endl;
        std::cout << "x:" << quat.at(1) << std::endl;
        std::cout << "y:" << quat.at(2) << std::endl;
        std::cout << "z:" << quat.at(3) << std::endl;
    }
}

void exampleQuatToRpy(RpcClientPtr impl)
{
    std::cout << std::endl << "四元数 -> 欧拉角" << std::endl;

    std::vector<double> quat;
    quat.push_back(0.789);
    quat.push_back(0.153);
    quat.push_back(0.001);
    quat.push_back(0.595);
    std::cout << "w:" << quat.at(0) << std::endl;
    std::cout << "x:" << quat.at(1) << std::endl;
    std::cout << "y:" << quat.at(2) << std::endl;
    std::cout << "z:" << quat.at(3) << std::endl;

    auto robot_name = impl->getRobotNames().front();
    std::vector<double> rpy = impl->getRobotInterface(robot_name)
                                  ->getRobotAlgorithm()
                                  ->quaternionToRpy(quat);

    if (3 == rpy.size()) {
        std::cout << "rx:" << rpy.at(0) << std::endl;
        std::cout << "ry:" << rpy.at(1) << std::endl;
        std::cout << "rz:" << rpy.at(2) << std::endl;
    }
}

std::tuple<std::vector<double>, int> res;

void exampleInverseK(RpcClientPtr impl)
{
    std::cout << std::endl << "逆解" << std::endl;
    auto robot_name = impl->getRobotNames().front();
    auto cur_joint = impl->getRobotInterface(robot_name)
                         ->getRobotState()
                         ->getJointPositions();

    //机械臂目标点
    std::vector<double> Waypoint_1_p = { 0.549273, -0.12094,   0.464501,
                                         3.13802,  0.00122674, 1.57095 };
    // 机械臂当前位置作为参考点
    std::vector<double> Waypoint_0_q = { 0.00123059, -0.26063, 1.7441,
                                         0.437503,   1.56957,  0.00107586 };

    res = impl->getRobotInterface(robot_name)
              ->getRobotAlgorithm()
              ->inverseKinematics(Waypoint_0_q, Waypoint_1_p);

    std::cout << "逆解位置" << std::endl;
    for (int i = 0; i < 6; i++) {
        std::cout << Waypoint_1_p.at(i) << ",";
    }
    std::cout << std::endl << "逆解结果:" << std::get<1>(res) << std::endl;
    for (int i = 0; i < 6; i++) {
        std::cout << std::get<0>(res).at(i) << ",";
    }
}

void exampleForwardK(RpcClientPtr impl)
{
    std::cout << std::endl << "正解" << std::endl;

    auto robot_name = impl->getRobotNames().front();
    auto res_pose = impl->getRobotInterface(robot_name)
                        ->getRobotAlgorithm()
                        ->forwardKinematics(std::get<0>(res));

    std::cout << "正解关节" << std::endl;
    for (int i = 0; i < 6; i++) {
        std::cout << std::get<0>(res).at(i) << ",";
    }
    std::cout << std::endl << "正解结果:" << std::get<1>(res_pose) << std::endl;
    for (int i = 0; i < 6; i++) {
        std::cout << std::get<0>(res_pose).at(i) << ",";
    }
}

#define LOCAL_IP "127.0.0.1"

int main(int argc, char **argv)
{
    auto rpc_cli = std::make_shared<RpcClient>();
    rpc_cli->connect(LOCAL_IP, 30004);
    rpc_cli->login("aubo", "123456");
    rpc_cli->setRequestTimeout(20); // pathMovej耗时

    exampleRpyToQuat(rpc_cli);
    exampleQuatToRpy(rpc_cli);
    exampleInverseK(rpc_cli);
    exampleForwardK(rpc_cli);

    auto robot_name = rpc_cli->getRobotNames().front();

    // calibrateTcpForceSensor
    std::vector<std::vector<double>> q0(6, { 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 });
    std::vector<std::vector<double>> qd(6, { 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 });
    std::vector<std::vector<double>> qdd(6, { 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 });
    std::vector<std::vector<double>> temp(
        6, { 30.1, 30.1, 30.1, 40.1, 40.1, 40.1 });
    auto result0 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->frictionModelIdentify(q0, qd, qdd, temp);

    // payloadIdentify
    std::vector<std::vector<double>> q(6, { 0., 0., 0., 0., 0., 0. });
    std::vector<std::vector<double>> forces(6, { 0., 0., 0., 0., 0., 0. });
    auto result1 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->payloadIdentify(q, forces);

    // frictionModelIdentify(舍弃)

    // tcpOffsetIdentify
    std::vector<std::vector<double>> tool_poses(
        { { 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 },
          { 0.2, 0.2, 0.2, 0.2, 0.2, 0.2 },
          { 0.3, 0.3, 0.3, 0.3, 0.3, 0.3 },
          { 0.4, 0.4, 0.4, 0.4, 0.4, 0.4 } });
    auto result2 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->tcpOffsetIdentify(tool_poses);

    // coordinateCalibrate
    std::vector<std::vector<double>> coord_poses(
        { { 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 },
          { 0.2, 0.2, 0.2, 0.2, 0.2, 0.2 },
          { 0.3, 0.3, 0.3, 0.3, 0.3, 0.3 },
          { 0.4, 0.4, 0.4, 0.4, 0.4, 0.4 } });
    /**
    auto result3 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->coordinateCalibrate(1, coord_poses);
    **/

    // eyeInhandCalibration
    std::vector<std::vector<double>> hand(6, { 0., 0., 0., 0., 0., 0. });
    std::vector<std::vector<double>> eye(6, { 0., 0., 0., 0., 0., 0. });
    auto result4 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->eyeInhandCalibration(hand, eye);

    // eyeTohandCalibration
    auto result5 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->eyeTohandCalibration(hand, eye);

    // calibWorkpieceCoordinatePara(工件标定)
    auto result6 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->calibWorkpieceCoordinatePara(coord_poses, 1);

    // forwardDynamics
    std::vector<double> pose(6, 0.);
    std::vector<double> torqs(6, 0);
    auto result7 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->forwardDynamics(pose, torqs);
    // poseTrans(位置变换)
    std::vector<double> pose_from(6, 0.);
    std::vector<double> pose_from_to(6, 0.5);
    auto result8 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->poseTrans(pose_from, pose_from_to);

    // poseInverse(齐次变换矩阵的逆)
    auto result9 = rpc_cli->getRobotInterface(robot_name)
                       ->getRobotAlgorithm()
                       ->poseInverse({ 0.1, 0.1, 0.1, 0.1, 0.1, 0.1 });

    // poseTransInv(求逆)
    std::vector<double> F_B_A(6, 0.);
    std::vector<double> F_C_A(6, 0.5);
    auto result10 = rpc_cli->getRobotInterface(robot_name)
                        ->getRobotAlgorithm()
                        ->poseTransInv(F_B_A, F_C_A);

    // poseRotation(姿态变换)
    std::vector<double> pose1(6, 0.1);
    std::vector<double> rotv(6, 0.2);
    auto result11 = rpc_cli->getRobotInterface(robot_name)
                        ->getRobotAlgorithm()
                        ->poseRotation(pose1, rotv);

    // pathMovej
    std::vector<double> q1(6, 0.);
    std::vector<double> q2(6, 1.1);
    auto result14 = rpc_cli->getRobotInterface(robot_name)
                        ->getRobotAlgorithm()
                        ->pathMovej(q1, 0.001, q2, 0.001, 0.02);

    // pathBlend3Points
    std::vector<double> q_start(6, 0.1);
    std::vector<double> q_via(6, 0.2);
    std::vector<double> q_to(6, 0.3);
    auto result15 = rpc_cli->getRobotInterface(robot_name)
                        ->getRobotAlgorithm()
                        ->pathBlend3Points(1, q_start, q_via, q_to, 0.01, 0.08);

    return 0;
}
